﻿using System;
using System.Diagnostics;
using System.IO;
using Microsoft.Web.Deployment;

namespace MSDeployCustomProviders
{
    // Usage: 
    // msdeploy -verb:sync -source:batch=d:\myBatchJob.bat -dest:auto
    public class BatchProvider : DeploymentObjectProvider
    {
        public BatchProvider(DeploymentProviderContext providerContext, DeploymentBaseContext baseContext)
            : base(providerContext, baseContext)
        {
        }

        #region Data Members

        internal const string ObjectName = "batch";
        internal const string KeyAttributeName = "path";

        protected string _filePath;

        #endregion

        #region Properties

        public override string Name
        {
            get { return ObjectName; }
        }

        private string Path
        {
            get
            {
                return this.ProviderContext.Path;
            }
        }

        #endregion

        #region Methods

        public override void GetAttributes(DeploymentAddAttributeContext addContext)
        {
            // always throw on destination object
            if (this.BaseContext.IsDestinationObject)
            {
                // Dummy exception to invoke add
                throw new Exception();
            }
            // on source object, check if the path is valid
            if (!File.Exists(this.Path))
            {
                throw new DeploymentFatalException("File <" + this.Path + "> does not exist");
            }
        }

        ///---------------------------------------------------------------
        /// <summary>
        /// Endows the provider with the attribute "path"
        /// </summary>
        ///---------------------------------------------------------------
        public override DeploymentObjectAttributeData CreateKeyAttributeData()
        {
            return new DeploymentObjectAttributeData(BatchProvider.KeyAttributeName,
                this.Path,
                DeploymentObjectAttributeKind.CaseInsensitiveCompare);
        }

        ///---------------------------------------------------------------
        /// <summary>
        /// This method is called by when syncing to an archive
        /// So that when packaging with package or archiveDir providers, 
        /// the actual physical file is included in the package/archive
        ///---------------------------------------------------------------
        public override Stream GetStream()
        {
            return new FileStream(Path, FileMode.Open, FileAccess.Read);
        }

        ///---------------------------------------------------------------
        /// <summary>
        /// Add is called by the destination object when performing a sync
        /// to validate the object (after GetAttributes throws)
        /// </summary>
        ///---------------------------------------------------------------
        public override void Add(DeploymentObject source, bool whatIf)
        {
            if (!whatIf && File.Exists(source.ProviderContext.Path))
            {
                ProcessStartInfo batchFileInfo = new ProcessStartInfo(source.ProviderContext.Path);
                batchFileInfo.CreateNoWindow = true;
                Process batchProcess = new Process();
                batchProcess.StartInfo = batchFileInfo;
                batchProcess.Start();
                batchProcess.WaitForExit();
            }
        }

        public override void Update(DeploymentObject source, bool whatIf)
        {
            this.Add(source, whatIf);
        }

        public override DeploymentObjectKind ObjectKind
        {
            get
            {
                return DeploymentObjectKind.ContainsStream;
            }
        }

        #endregion
    }
}
